# This file is part of aiootp, an asynchronous one-time-pad based crypto
# and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2020 Gonzo Investigatory Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2020 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = [
    "keygens",
    "AsyncKeys",
    "Keys",
    "amnemonic",
    "mnemonic",
    "akeypair",
    "keypair",
]


__doc__ = """
A collection of highlevel tools for creating & managing symmetric keys.
"""


from .asynchs import *
from .commons import *
from .randoms import salt
from .randoms import asalt
from .randoms import csprbg
from .randoms import csprng
from .randoms import acsprng
from .randoms import token_bytes
from .randoms import atoken_bytes
from .ciphers import keys
from .ciphers import akeys
from .ciphers import Passcrypt
from .ciphers import passcrypt
from .ciphers import apasscrypt
from .ciphers import bytes_keys
from .ciphers import abytes_keys
from .ciphers import OneTimePad
from .ciphers import keypair_ratchets
from .ciphers import akeypair_ratchets
from .generics import azip
from .generics import is_iterable
from .generics import comprehension
from .generics import sha_256
from .generics import asha_256
from .generics import sha_256_hmac
from .generics import asha_256_hmac
from .generics import convert_static_method_to_member


@comprehension()
async def atable_key_gen(key=None, table=ASCII_TABLE):
    """
    This table based key generator function converts any key string
    containing an arbitrary set of characters, into another key string
    containing the set of characters provided by the table argument.
    This is an infinite generator that produces the elements one at a
    time.

    The ASCII_TABLE that's provided as a default, is a comprehensive set
    of ascii characters that are legible, unique, and have single octet
    byte representations. This ensures keys can be converted into a
    cryptographically secure byte stream of numbers from 32 to
    126.

    This generator function provides either deterministic keys from a
    user key, or generates a random 512-bit hash and derives a random
    key with the desired table elements from this hash. The result is a
    random, normal distribution of characters from among the items
    within the table.

    Usage Examples:

    # To produce a 60 byte key of characters from the default table
    key = "hotdiggitydog_thischowisyummy"
    async with atable_key_gen(key=key) as generator:
        new_key = await generator()
        assert new_key != await generator()
    print(new_key)
    >>> Hx`4^ej;u&/]qOF21Ea2~(6f"smp'DvMk[(wy'lME%CpCo|1ZWt> &tu=Mw_
    """
    if not key:
        key = await acsprng()
    size = len(table[:])
    async for key_portion in akeys(key, key).aint(16).ato_base(size, table):
        yield key_portion


@comprehension()
def table_key_gen(key=None, table=ASCII_TABLE):
    """
    This table based key generator function converts any key string
    containing an arbitrary set of characters, into another key string
    containing the set of characters provided by the table argument.
    This is an infinite generator that produces the elements one at a
    time.

    The ASCII_TABLE that's provided as a default, is a comprehensive set
    of ascii characters that are legible, unique, and have single octet
    byte representations. This ensures keys can be converted into a
    cryptographically secure byte stream of numbers from 32 to
    126.

    This generator function provides either deterministic keys from a
    user key, or generates a random 512-bit hash and derives a random
    key with the desired table elements from this hash. The result is a
    random, normal distribution of characters from among the items
    within the table.

    Usage Example:
    key = "hotdiggitydog_thischowisyummy"
    with table_key_gen(key=key) as generator:
        new_key = generator()
        assert new_key != generator()
    print(new_key)
    >>> Hx`4^ej;u&/]qOF21Ea2~(6f"smp'DvMk[(wy'lME%CpCo|1ZWt> &tu=Mw_
    """
    if not key:
        key = csprng()
    size = len(table[:])
    for key_portion in keys(key, key).int(16).to_base(size, table):
        yield key_portion


async def atable_key(key=None, table=ASCII_TABLE, size=64):
    """
    This table based key function converts any key string containing
    any arbitrary set of characters, into another key string containing
    the set of items provided by the table argument.

    The ASCII_TABLE that's provided as a default, is a comprehensive set
    of ascii characters that are all legible, with unique, single octet
    byte representations. This ensures keys can be converted into byte
    streams of numbers from 32 to 126, with no duplicate values.

    This generator function provides either deterministic keys from a
    user key, or generates a random 512-bit hash and derives a random
    key with the desired table elements from this hash. The result is a
    random, normal distribution of characters from among the items
    within the table.

    The size parameter determines the number of bytes/elements the
    output will contain.


    Usage Examples:

    key = "smellycaaaat, smelly caaaaat!"
    new_key = table_key(key=key, table="0123456789abcdef")
    print(new_key)
    >>> 4f271c61b0e615a7d3e9ac0161497034d047d4ecddc650ae054f829b3416818c

    new_key = table_key(key=key, size=len(key))
    print(new_key)
    >>> #mE)bOQD@lY%]Qwpb9Zi^32]jteVg
    """
    async with atable_key_gen(key=key, table=table) as generator:
        new_key = await generator()
        while len(new_key) < size:
            new_key += await generator()
        return new_key[:size]


def table_key(key=None, table=ASCII_TABLE, size=64):
    """
    This table based key function converts any key string containing
    any arbitrary set of characters, into another key string containing
    the set of items provided by the table argument.

    The ASCII_TABLE that's provided as a default, is a comprehensive set
    of ascii characters that are all legible, with unique, single octet
    byte representations. This ensures keys can be converted into byte
    streams of numbers from 32 to 126, with no duplicate values.

    This generator function provides either deterministic keys from a
    user key, or generates a random 512-bit hash and derives a random
    key with the desired table elements from this hash. The result is a
    random, normal distribution of characters from among the items
    within the table.

    The size parameter determines the number of bytes/elements the
    output will contain.


    Usage Examples:
    key = "smellycaaaat, smelly caaaaat!"
    new_key = table_key(key=key, table="0123456789abcdef")
    print(new_key)
    >>> 4f271c61b0e615a7d3e9ac0161497034d047d4ecddc650ae054f829b3416818c

    new_key = table_key(key=key, size=len(key))
    print(new_key)
    >>> #mE)bOQD@lY%]Qwpb9Zi^32]jteVg
    """
    with table_key_gen(key=key, table=table) as generator:
        new_key = generator()
        while len(new_key) < size:
            new_key += generator()
        return new_key[:size]


@comprehension()
async def amnemonic(key, salt=None, words=WORD_LIST):
    """
    Creates a stream of words for a mnemonic key from a user password
    ``key`` & random salt. If a salt isn't passed, then a random salt is
    generated & is available by calling ``result(exit=True)`` on the
    generator object. The ``words`` used for the mnemonic can be passed
    in, but by default are a 2048 word list of unique, all lowercase
    english words.
    """
    translate = None
    length = len(words)
    salt = salt if salt else await acsprng()
    key = await apasscrypt(key, salt)
    entropy = abytes_keys(key, salt, key)
    async with entropy.abytes_to_int().arelay(salt) as indexes:
        while True:
            if translate:
                await entropy.gen.asend(translate)
            translate = yield words[await indexes() % length]


@comprehension()
def mnemonic(key, salt=None, words=WORD_LIST):
    """
    Creates a stream of words for a mnemonic key from a user password
    ``key`` & random salt. If a salt isn't passed, then a random salt is
    generated & is available by calling ``result(exit=True)`` on the
    generator object. The ``words`` used for the mnemonic can be passed
    in, but by default are a 2048 word list of unique, all lowercase
    english words.
    """
    translate = None
    length = len(words)
    salt = salt if salt else csprng()
    key = passcrypt(key, salt)
    entropy = bytes_keys(key, salt, key)
    with entropy.bytes_to_int().relay(salt) as indexes:
        while True:
            if translate:
                entropy.gen.send(translate)
            translate = yield words[indexes() % length]


async def akeypair(entropy=csprbg()):
    """
    Returns a pair of symmetric hexidecimal keys from our fast
    cryptographically secure pseudo-random number generator. One is a
    512-bit encryption key, the other is a 256-bit salt.
    """
    return commons.Namespace(
        key=await acsprng(entropy), salt=await asalt(entropy)
    )


def keypair(entropy=csprbg()):
    """
    Returns a pair of symmetric hexidecimal keys from our fast
    cryptographically secure pseudo-random number generator. One is a
    512-bit encryption key, the other is a 256-bit salt.
    """
    return commons.Namespace(key=csprng(entropy), salt=salt(entropy))


class AsyncKeys:
    """
    This simple class coordinates and manages a symmetric key for
    establishing an arbitrary number of secure, deterministic streams of
    key material through an instance's ``__getitem__`` method.
    The class also contains static method key generators which function
    independantly from instance states, as well as the ability to create
    & validate HMAC code.
    """

    instance_methods = {
        akeys, abytes_keys, amnemonic, atable_key, atable_key_gen
    }

    asalt = staticmethod(asalt)
    akeys = staticmethod(akeys)
    akeypair = staticmethod(akeypair)
    amnemonic = staticmethod(amnemonic)
    apasscrypt = staticmethod(apasscrypt)
    abytes_keys = staticmethod(abytes_keys)
    atable_key = staticmethod(atable_key)
    atable_key_gen = staticmethod(atable_key_gen)
    akeypair_ratchets = staticmethod(akeypair_ratchets)

    def __init__(self, key=None):
        """
        Stores a key in the instance used to create deterministic
        streams of key material &, create & validate HMAC codes. If a
        ``key`` argument is not passed then a new 512-bit random key is
        created.
        """
        self._key = key if key else salt()
        for method in self.instance_methods:
            convert_static_method_to_member(
                self, method.__name__, method, key=self.key,
            )
        self.apasscrypt = self._apasscrypt

    def __getitem__(self, pid=""):
        """
        Provides a simple interface for users to create deterministic
        & externally uncorrelatable key material stream from a user-
        defined ``pid`` value.

        Usage Example:

        import aiootp

        keyring = aiootp.AsyncKeys(key)
        keystream = keyring["conversation"]
        datastream = aiootp.adata("Hey, when's the party?").aascii_to_int()

        async with datastream.axor(keystream) as ciphering:
            ciphered = await ciphering.alist()

        salt = await keystream.aresult(exit=True)

        derived_keystream = keyring.akeys(salt=salt, pid="conversation")
        deciphering = aiootp.aunpack(ciphered).axor(derived_keystream)

        async with deciphering.aint_to_ascii() as plaintext:
            assert "Hey, when's the party?" == await plaintext.ajoin()
        """
        return akeys(key=self.key, pid=pid)

    @property
    def key(self):
        """
        Returns the instance's main symmetric key.
        """
        return self._key

    async def ahmac(
        self, data=None, *, key=None, hasher=asha_256_hmac
    ):
        """
        Creates an HMAC code of ``data`` using ``key``  or the
        instance's ``self.key`` if it's not supplied & the hashing
        function ``hasher``.
        """
        return await hasher(data, key=key if key else self.key)

    async def atest_hmac(
        self, data=None, hmac=None, *, key=None, hasher=asha_256_hmac
    ):
        """
        Tests if the given ``hmac`` of some ``data`` is valid with a
        non-constant time comparison on the hash of each the supplied &
        derived HMACs, appended with a salt prior to hashing. The
        algorithm prepends the instance's ``self.key`` if ``key`` is not
        supplied to further make the tested outputs undeterminable to an
        attacker. The random salt & key allow the hashes to be compared
        normally in non-constant time, without revealing meaningful
        information, since an attacker wouldn't have access to either.
        This scheme is easier to implement correctly & is easier to
        prove guarantees of the infeasibility of timing attacks. Any
        async ``hasher`` function can be specified as the HMAC function,
        which is by default ``asha_256_hmac``.
        """
        if not hmac:
            raise ValueError("``hmac`` argument was not given.")
        key = key if key else self.key
        true_hmac = await self.ahmac(data=data, key=key, hasher=hasher)
        if await self.atime_safe_equality(hmac, true_hmac, key=key):
            return True
        else:
            raise ValueError("HMAC of ``data`` isn't valid.")

    async def atime_safe_equality(
        self, value_0=None, value_1=None, *, key=None
    ):
        """
        Tests if ``value_0`` is equal to ``value_1`` with a non-constant
        time comparison on the hash of each value appended with a salt
        prior to hashing. The algorithm prepends the instance's
        ``self.key`` if ``key`` is not supplied to further make the
        tested outputs undeterminable to an attacker. The random salt
        & key allow the hashes to be compared normally in non-constant
        time, without revealing meaningful information, since an
        attacker wouldn't have access to either. This scheme is easier
        to implement correctly & is easier to prove guarantees of the
        infeasibility of timing attacks.
        """
        salt = await atoken_bytes(32)
        key = key if key else self.key
        if (
            await asha_256(key, value_0, salt)
            == await asha_256(key, value_1, salt)
        ):
            return True
        else:
            return False

    async def areset(self, key=None):
        """
        Replaces the stored instance key used to create deterministic
        streams of key material &, create & validate HMAC codes.
        """
        self._key = key if key else await acsprng()
        for method in self.instance_methods:
            convert_static_method_to_member(
                self, method.__name__, method, key=self.key,
            )
            await switch()

    async def _apasscrypt(
        self, password, salt, *, kb=1024, cpu=3, hardness=1024
    ):
        """
        A tunably memory & cpu hard method which returns a key from a
        user password & salt. This method also protects the passwords
        it processes with the instance's key, which forces attackers to
        also find a way to retrieve it in order to crack the passwords.
        """
        Passcrypt._check_inputs(password, salt)
        salted_password = await self.ahmac((password, salt))
        return await apasscrypt(
            salted_password, salt, kb=kb, cpu=cpu, hardness=hardness
        )


class Keys:
    """
    This simple class coordinates and manages a symmetric key for
    establishing an arbitrary number of secure, deterministic streams of
    key material through an instance's ``__getitem__`` method.
    The class also contains static method key generators which function
    independantly from instance states, as well as the ability to create
    & validate HMAC code.
    """

    instance_methods = {
        keys, bytes_keys, mnemonic, table_key, table_key_gen
    }

    salt = staticmethod(salt)
    keys = staticmethod(keys)
    keypair = staticmethod(keypair)
    mnemonic = staticmethod(mnemonic)
    passcrypt = staticmethod(passcrypt)
    bytes_keys = staticmethod(bytes_keys)
    table_key = staticmethod(table_key)
    table_key_gen = staticmethod(table_key_gen)
    keypair_ratchets = staticmethod(keypair_ratchets)

    def __init__(self, key=None):
        """
        Stores a key in the instance used to create deterministic
        streams of key material &, create & validate HMAC codes. If a
        ``key`` argument is not passed then a new 512-bit random key is
        created.
        """
        self.reset(key)
        self.passcrypt = self._passcrypt

    def __getitem__(self, pid=""):
        """
        Provides a simple interface for users to create deterministic
        & externally uncorrelatable key material stream from a user-
        defined ``pid`` value.

        Usage Example:

        import aiootp

        keyring = aiootp.Keys(key)
        keystream = keyring["conversation"]
        datastream = aiootp.data("Hey, when's the party?").ascii_to_int()

        with datastream.xor(keystream) as ciphering:
            ciphered = ciphering.list()

        salt = keystream.result(exit=True)

        derived_keystream = keyring.keys(salt=salt, pid="conversation")
        deciphering = aiootp.unpack(ciphered).xor(derived_keystream)

        with deciphering.int_to_ascii() as plaintext:
            assert "Hey, when's the party?" == plaintext.join()
        """
        return keys(key=self.key, pid=pid)

    @property
    def key(self):
        """
        Returns the instance's main symmetric key.
        """
        return self._key

    def hmac(self, data=None, *, key=None, hasher=sha_256_hmac):
        """
        Creates an HMAC code of ``data`` using ``key``  or the
        instance's ``self.key`` if it's not supplied & the hashing
        function ``hasher``.
        """
        return hasher(data, key=key if key else self.key)

    def test_hmac(
        self, data=None, hmac=None, *, key=None, hasher=sha_256_hmac
    ):
        """
        Tests if the given ``hmac`` of some ``data`` is valid with a
        non-constant time comparison on the hash of each the supplied &
        derived HMACs, appended with a salt prior to hashing. The
        algorithm prepends the instance's ``self.key`` if ``key`` is not
        supplied to further make the tested outputs undeterminable to an
        attacker. The random salt & key allow the hashes to be compared
        normally in non-constant time, without revealing meaningful
        information, since an attacker wouldn't have access to either.
        This scheme is easier to implement correctly & is easier to
        prove guarantees of the infeasibility of timing attacks. Any
        sync ``hasher`` function can be specified as the HMAC function,
        which is by default ``sha_256_hmac``.
        """
        if not hmac:
            raise ValueError("``hmac`` argument was not given.")
        key = key if key else self.key
        true_hmac = self.hmac(data=data, key=key, hasher=hasher)
        if self.time_safe_equality(hmac, true_hmac, key=key):
            return True
        else:
            raise ValueError("HMAC of ``data`` isn't valid.")

    def time_safe_equality(self, value_0=None, value_1=None, *, key=None):
        """
        Tests if ``value_0`` is equal to ``value_1`` with a non-constant
        time comparison on the hash of each value appended with a salt
        prior to hashing. The algorithm prepends the instance's
        ``self.key`` if ``key`` is not supplied to further make the
        tested outputs undeterminable to an attacker. The random salt
        & key allow the hashes to be compared normally in non-constant
        time, without revealing meaningful information, since an
        attacker wouldn't have access to either. This scheme is easier
        to implement correctly & is easier to prove guarantees of the
        infeasibility of timing attacks.
        """
        salt = token_bytes(32)
        key = key if key else self.key
        if sha_256(key, value_0, salt) == sha_256(key, value_1, salt):
            return True
        else:
            return False

    def reset(self, key=None):
        """
        Replaces the stored instance key used to create deterministic
        streams of key material &, create & validate HMAC codes.
        """
        self._key = key if key else csprng()
        for method in self.instance_methods:
            convert_static_method_to_member(
                self, method.__name__, method, key=self.key,
            )

    def _passcrypt(
        self, password, salt, *, kb=1024, cpu=3, hardness=1024
    ):
        """
        A tunably memory & cpu hard method which returns a key from a
        user password & salt. This method also protects the passwords
        it processes with the instance's key, which forces attackers to
        also find a way to retrieve it in order to crack the passwords.
        """
        Passcrypt._check_inputs(password, salt)
        salted_password = self.hmac((password, salt))
        return passcrypt(
            salted_password, salt, kb=kb, cpu=cpu, hardness=hardness
        )


async def ainsert_keyrings(self, key=None):
    """
    A generic __init__ function that can be copied into abitrary class
    or instance dictionaries to give those objects access to stateful
    & ephemeral key material generators.
    """
    key = key if key else await acsprng()
    self.keyring = Keys(key=key)
    self.akeyring = AsyncKeys(key=key)


def insert_keyrings(self, key=None):
    """
    A generic __init__ function that can be copied into abitrary class
    or instance dictionaries to give those objects access to stateful
    & ephemeral key material generators.
    """
    key = key if key else csprng()
    self.keyring = Keys(key=key)
    self.akeyring = AsyncKeys(key=key)


__extras = {
    "AsyncKeys": AsyncKeys,
    "Keys": Keys,
    "__doc__": __doc__,
    "__main_exports__": __all__,
    "__package__": "aiootp",
    "ainsert_keyrings": ainsert_keyrings,
    "insert_keyrings": insert_keyrings,
    "akeys": akeys,
    "keys": keys,
    "mnemonic": mnemonic,
    "akeypair": akeypair,
    "keypair": keypair,
    "akeypair_ratchets": akeypair_ratchets,
    "keypair_ratchets": keypair_ratchets,
    "amnemonic": amnemonic,
    "atable_key": atable_key,
    "table_key": table_key,
    "atable_key_gen": atable_key_gen,
    "table_key_gen": table_key_gen,
}


keygens = Namespace.make_module("keygens", mapping=__extras)




# This file is part of aiootp, an asynchronous one-time-pad based crypto
# and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2020 Gonzo Investigatory Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2020 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = ["paths", "SecurePath", "DatabasePath"]


__doc__ = """
A small collection of ``pathlib.Path`` objects with references to
potentially helpful directories.
"""


import os
import aiofiles
from pathlib import Path
from .asynchs import aos
from .asynchs import switch
from .commons import Namespace


def CurrentPath():
    """
    Returns a ``pathlib.Path`` object pointing to the current working
    directory.
    """
    return Path(os.getcwd()).absolute()


def RootPath():
    """
    Returns a ``pathlib.Path`` object pointing to this module's directory.
    """
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return Path(dir_path).absolute()


def PackagePath():
    """
    Returns a ``pathlib.Path`` object pointing to this package's top
    level directory.
    """
    return RootPath().parent


def TorPath(dir_function=RootPath):
    """
    Returns a ``pathlib.Path`` object pointing to an optional tor
    directory.
    """
    return dir_function() / "tor"


def DatabasePath(dir_function=RootPath):
    """
    Returns a ``pathlib.Path`` object pointing to the default directory
    encrypted databases are saved to.
    """
    return dir_function() / "databases"


async def amake_hash_file(path):
    from .ciphers import asalt

    key = await asalt()
    filename = key[:64]
    secret = key[64:]
    filepath = path / filename
    async with aiofiles.open(filepath, "w") as new_file:
        await new_file.write(secret)
    await aos.chmod(filepath, 0o000)


def make_hash_file(path):
    from .ciphers import salt

    key = salt()
    filename = key[:64]
    secret = key[64:]
    filepath = path / filename
    with open(filepath, "w") as new_file:
        new_file.write(secret)
    os.chmod(filepath, 0o000)


async def afind_hash_file(path):
    for subpath in path.iterdir():
        if subpath.is_file() and len(subpath.stem) == 64:
            return subpath.absolute()
        await switch()


def find_hash_file(path):
    for subpath in path.iterdir():
        if subpath.is_file() and len(subpath.stem) == 64:
            return subpath.absolute()


async def aread_hash_file(filepath):
    try:
        await aos.chmod(filepath, 0o700)
        async with aiofiles.open(filepath, "r") as hash_file:
            return await hash_file.read()
    finally:
        await aos.chmod(filepath, 0o000)


def read_hash_file(filepath):
    try:
        os.chmod(filepath, 0o700)
        with open(filepath, "r") as hash_file:
            return hash_file.read()
    finally:
        os.chmod(filepath, 0o000)


def SecurePath(dir_function=DatabasePath):
    path = dir_function() / "secure"
    if not path.exists():
        path.mkdir()
    filepath = find_hash_file(path)
    if not filepath:
        make_hash_file(path)
        return find_hash_file(path)
    else:
        return filepath


__extras = {
    "Path": Path,
    "TorPath": TorPath,
    "RootPath": RootPath,
    "SecurePath": SecurePath,
    "PackagePath": PackagePath,
    "CurrentPath": CurrentPath,
    "DatabasePath": DatabasePath,
    "__doc__": __doc__,
    "__main_exports__": __all__,
    "__package__": "aiootp",
    "_afind_hash_file": afind_hash_file,
    "_find_hash_file": find_hash_file,
    "_amake_hash_file": amake_hash_file,
    "_make_hash_file": make_hash_file,
    "_aread_hash_file": aread_hash_file,
    "_read_hash_file": read_hash_file,
}


paths = Namespace.make_module("paths", mapping=__extras)




# This file is part of tiny_onion, a small-as-possible solution for p2p
# networking over tor v3 onion services.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2020 Gonzo Investigatory Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#          © 2019-2020 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


from init_tests import *


__all__ = [
    "test_comprehension_context",
    "test_comprehension_iter",
    "test_chainable_methods",
    "test_comprehension",
    "__all__",
]


TEST_STRING = "abcdefghijk"
TEST_STRING_LENGTH = len(TEST_STRING)
assert TEST_STRING_LENGTH == 11


@comprehension()
async def awith_return_coro(*, got=None):
    for _ in range(TEST_STRING_LENGTH):
        got = yield got
        await switch()
    raise UserWarning(got)


@comprehension()
def with_return_coro(*, got=None):
    for _ in range(TEST_STRING_LENGTH):
        got = yield got
    return got


@comprehension()
async def ano_return_coro(*, got=None):
    for _ in range(TEST_STRING_LENGTH):
        await switch()
        got = yield got


@comprehension()
def no_return_coro(*, got=None):
    for _ in range(TEST_STRING_LENGTH):
        got = yield got


@comprehension()
async def awith_return_iterator():
    for char in TEST_STRING:
        await switch()
        yield char
    raise UserWarning(char)


@comprehension()
def with_return_iterator():
    for char in TEST_STRING:
        yield char
    return char


@comprehension()
async def ano_return_iterator():
    for char in TEST_STRING:
        await switch()
        yield char


@comprehension()
def no_return_iterator(got=TEST_STRING):
    for char in TEST_STRING:
        yield char


def test_comprehension():
    """
    Tests the ``comprehension`` decorator which wraps generator
    functions so they return ``Comprende`` objects when the function
    is called to create an async/sync generator.
    """
    @comprehension()
    def is_Comprende_generator():
        yield

    assert is_Comprende_generator().__class__ == Comprende

    @comprehension()
    async def is_async_Comprende_generator():
        yield

    assert is_async_Comprende_generator().__class__ == Comprende

    async def run_test_generators():
        for value in is_Comprende_generator():
            assert value == None

        async for value in is_async_Comprende_generator():
            assert value == None

    run(run_test_generators())


def test_comprehension_context():
    """
    Testing the accuracy of async/sync ``Comprende``  generator results
    and the class' ability to return values from completed generators.
    """
    run(comprehension_context_testing())


async def comprehension_context_testing():
    """
    Testing the accuracy of async/sync ``Comprende``  generator results
    and the class' ability to return values from completed generators.
    """
    async with awith_return_coro() as asend:
        assert await asend(None) == None
        for test in TEST_STRING:
            assert await asend(test) == test
    assert await asend.aresult() == "k", await asend.aresult()

    with with_return_coro() as send:
        assert send(None) == None
        for test in TEST_STRING:
            assert send(test) == test
    assert send.result() == "k", send.result()

    async with ano_return_coro() as asend:
        assert await asend(None) == None
        for test in TEST_STRING:
            assert await asend(test) == test
    try:
        failed = False
        return_value = await asend.aresult()
        assert return_value != None
    except AssertionError:
        failed = True
    finally:
        assert failed, return_value

    with no_return_coro() as send:
        assert send(None) == None
        for test in TEST_STRING:
            assert send(test) == test
    try:
        failed = False
        return_value = send.result()
        assert return_value != None
    except AssertionError:
        failed = True
    finally:
        assert failed, return_value


def test_comprehension_iter():
    """
    Testing the accuracy of async/sync ``Comprende``  generator results
    and the class' ability to return values from completed generators.
    """
    run(comprehension_iteration_testing())


async def comprehension_iteration_testing():
    """
    Testing the accuracy of async/sync ``Comprende``  generator results
    and the class' ability to return values from completed generators.
    """
    catcher = awith_return_iterator()
    async for index, value in catcher.atag():
        assert value == TEST_STRING[index], value
    return_value = await catcher.aresult()
    assert return_value == "k", return_value

    catcher = with_return_iterator()
    for index, value in catcher.tag():
        assert value == TEST_STRING[index], value
    return_value = catcher.result()
    assert return_value == "k", return_value

    catcher = ano_return_iterator()
    async for index, value in catcher.atag():
        assert value == TEST_STRING[index], value
    try:
        failed = False
        return_value = await catcher.aresult()
        assert return_value != None
    except AssertionError:
        failed = True
    finally:
        assert failed, return_value

    catcher = no_return_iterator()
    for index, value in catcher.tag():
        assert value == TEST_STRING[index], value
    try:
        failed = False
        return_value = catcher.result()
        assert return_value != None
    except AssertionError:
        failed = True
    finally:
        assert failed, return_value


def test_chainable_methods():
    """
    Testing the chainable generator methods of the ``Comprende`` class.
    """
    run(Comprende_chainable_methods())


async def Comprende_chainable_methods():
    """
    Testing the chainable generator methods of the ``Comprende`` class.
    """
    # Check timeout / atimeout
    time_limit = 0.01
    time_start = asynchs.time()
    async for item in generics.arange(100_000_000).atimeout(time_limit):
        pass
    time_elapsed = asynchs.time() - time_start
    assert time_elapsed >= time_limit
    assert time_elapsed < time_limit + 0.001

    time_start = asynchs.time()
    for item in generics.range(100_000_000).timeout(time_limit):
        pass
    time_elapsed = asynchs.time() - time_start
    assert time_elapsed >= time_limit
    assert time_elapsed < time_limit + 0.001

    # Check halt / ahalt
    sentinel = csprng()[:2]
    chars = keys(key, salt, pid)
    achars = akeys(key, salt, pid)
    with keys(key, salt, pid)[:128] as sequence:
        sentinel_check = sequence.join()
    async for letters in achars.aresize(2).ahalt(sentinel):
        pass
    assert letters != sentinel
    assert letters + sentinel in sentinel_check

    for letters in chars.resize(2).halt(sentinel):
        pass
    assert letters != sentinel
    assert letters + sentinel in sentinel_check

    # Check feed / afeed
    mock_keys = keys(key, salt, pid)
    mock_food = order([None], range(4))
    async with akeys(key, salt, pid).afeed(range(5)) as altered_keys:
        first_loop = True
        for original_key in keys(key, salt, pid):
            altered_key = await altered_keys()
            if first_loop:
                first_loop = False
                assert original_key == altered_key
            else:
                assert original_key != altered_key
            with generics.ignore(StopIteration):
                assert mock_keys(mock_food()) == altered_key

    mock_keys = keys(key, salt, pid)
    mock_food = order([None], range(4))
    with keys(key, salt, pid).feed(range(5)) as altered_keys:
        first_loop = True
        async for original_key in akeys(key, salt, pid):
            altered_key = altered_keys()
            if first_loop:
                first_loop = False
                assert original_key == altered_key
            else:
                assert original_key != altered_key
            with generics.ignore(StopIteration):
                assert mock_keys(mock_food()) == altered_key

    # Check feed_self / afeed_self
    mock_keys = keys(key, salt, pid)
    async with akeys(key, salt, pid).afeed_self()[:5] as altered_keys:
        first_loop = True
        for original_key in keys(key, salt, pid):
            altered_key = await altered_keys()
            if first_loop:
                food = None
                first_loop = False
                assert original_key == altered_key
            else:
                assert original_key != altered_key
            with generics.ignore(StopIteration):
                food = mock_keys(food)
                assert food == altered_key

    mock_keys = keys(key, salt, pid)
    with keys(key, salt, pid).feed_self()[:5] as altered_keys:
        first_loop = True
        async for original_key in akeys(key, salt, pid):
            altered_key = altered_keys()
            if first_loop:
                food = None
                first_loop = False
                assert original_key == altered_key
            else:
                assert original_key != altered_key
            with generics.ignore(StopIteration):
                food = mock_keys(food)
                assert food == altered_key

    # Check heappop / aheappop
    mock_tags = generics.range(5)
    mock_keys = keys(key, salt, pid)
    async for tag, hex_number in akeys(key, salt, pid).atag().aheappop(5):
        assert tag == mock_tags()
        assert hex_number == mock_keys()

    mock_tags = generics.range(5)
    mock_keys = keys(key, salt, pid)
    async for hex_number, tag in akeys(key, salt, pid).aheappop(of=range(5)):
        assert tag == mock_tags()
        assert hex_number == mock_keys()

    mock_tags = generics.range(5)
    mock_keys = keys(key, salt, pid)
    for tag, hex_number in keys(key, salt, pid).tag().heappop(5):
        assert tag == mock_tags()
        assert hex_number == mock_keys()

    mock_tags = generics.range(5)
    mock_keys = keys(key, salt, pid)
    for hex_number, tag in keys(key, salt, pid).heappop(of=range(5)):
        assert tag == mock_tags()
        assert hex_number == mock_keys()

    # Check reversed / areversed
    reversed_list = list(reversed(range(5, 10)))
    async for tag, item in reversed(generics.arange(5, 10)).atag():
        assert item == reversed_list[tag]

    async for item, tag in generics.arange(5).areversed(of=range(5, 10)):
        assert tag == reversed_list[item]

    for tag, item in reversed(generics.range(5, 10)).tag():
        assert item == reversed_list[tag]

    for item, tag in generics.range(5).reversed(of=range(5, 10)):
        assert tag == reversed_list[item]

    # Check sort / asort
    tags = generics.arange(5)
    labels = akeys(key, salt, pid)
    mock_tags = generics.arange(5)
    mock_labels = akeys(key, salt, pid).aheappop(5)
    async for tag, label in tags.asort(span=5, of=labels):
        assert tag == await mock_tags()
        assert label == await mock_labels()

    tags = generics.range(5)
    labels = keys(key, salt, pid)
    mock_tags = generics.range(5)
    mock_labels = keys(key, salt, pid).heappop(5)
    for tag, label in tags.sort(span=5, of=labels):
        assert tag == mock_tags()
        assert label == mock_labels()

